{-# OPTIONS_GHC -fglasgow-exts -fbang-patterns #-}
module Sim
    ( Params
    , HiddenParams (..)
    , State
    , HiddenState (..)
    , simulate
    ) where

import Data.Complex

import Message
import Command

type Params = (InitParams, HiddenParams)
data HiddenParams
    = HiddenParams
    { accel :: !Int
    , brake :: !Int
    , rotAccel :: !Int
    }

type State = (VehicleState, HiddenState)
data HiddenState
    = HiddenState
    { rotSpeed :: !Double
    }

{-
「将来効果の現れるコマンドとその時間」のリストを渡して、指定した時間のシミュレート
処理されなかったコマンドのリストも返す。
-}
simulate :: Params -> State -> [(Time, Command)] ->
            Time -> Time -> (State, [(Time, Command)])
simulate !p !s [] !beg !end = (simulateSimple p s (end - beg), [])
simulate !p !s cs0@((t',cmd) : cs) !beg !end
    | t' < beg  = simulate p s cs beg end
    | t' < end  = simulate p s' cs t' end
    | otherwise = (simulateSimple p s (end - beg), cs0)
  where
    s' = updateControl cmd (simulateSimple p s (t' - beg))

simulateSimple :: Params -> State -> Time -> State
simulateSimple
  ( InitParams
    { maxSpeed = maxSpeed
    , maxTurn = maxTurn
    , maxHardTurn = maxHardTurn
    }
  , HiddenParams
    { accel = accel
    , brake = brake
    , rotAccel = rotAccel
    }
  )
  ( VehicleState
    { vehicleCtl = ctl@(ctlA, ctlT)
    , vehicleX = x
    , vehicleY = y
    , vehicleDir = dir
    , vehicleSpeed = s
    }
  , HiddenState
    { rotSpeed = rs
    }
  ) 
  t
  = (vs, hs)
    where
      dt = fromIntegral t / 1000 -- 秒単位へ
      pos = x :+ y
      pos' = pos + mkPolar (s*dt) (dir * pi / 180)
      vs = VehicleState
           { vehicleCtl = ctl
           , vehicleX = realPart pos'
           , vehicleY = imagPart pos'
           , vehicleDir = dir + rs * dt
           , vehicleSpeed =
               let !k = fromIntegral accel / (maxSpeed*maxSpeed)
                   !a = fromIntegral $ case ctlA of
                         Accelerating -> accel
                         Braking -> - brake
                         Rolling -> 0
               in max 0 $ s + dt*(a - k*s*s)
               -- s_{t'} = max(s_t + (t'-t)a - k(t'-t)s_{t}^2, 0)
           }
      hs = HiddenState
           { rotSpeed =
                 case ctlT of
                   HL -> min (rs + fromIntegral rotAccel * dt) maxTurn
                   L  -> min (rs + fromIntegral rotAccel * dt) maxHardTurn
                   R  -> max (rs - fromIntegral rotAccel * dt) (-maxTurn)
                   HR -> max (rs - fromIntegral rotAccel * dt) (-maxHardTurn)
                   ST -> rs
           }

updateControl :: Command -> State -> State
updateControl (cmdA, cmdT) (vs, hs) = seq vs' $ (vs', hs)
  where
    !(ctlA, ctlT) = vehicleCtl vs
    ctlA' = case cmdA of
              Nothing -> ctlA
              Just CmdA ->
                  case ctlA of
                    Accelerating -> Accelerating
                    Braking -> Rolling
                    Rolling -> Accelerating
              Just CmdB ->
                  case ctlA of
                    Accelerating -> Rolling
                    Braking -> Braking
                    Rolling -> Braking
    ctlT' = case cmdT of
              Nothing -> ctlT
              Just CmdL ->
                  case ctlT of
                    HL -> HL
                    L  -> HL
                    ST -> L
                    R  -> ST
                    HR -> R
              Just CmdR ->
                  case ctlT of
                    HL -> L
                    L  -> ST
                    ST -> R
                    R  -> HR
                    HR -> HR
    vs' = vs
          { vehicleCtl = (ctlA', ctlT')
          }
